@@ -1,24 +1,25 @@
 use v6;
-use Config::TOML::Parser::Exceptions;
+use Crane;
+use X::Config::TOML;
 unit class Config::TOML::Parser::Actions;
 
 # TOML document
-has %.toml;
+has %!toml;
 
 # TOML arraytable tracker, records arraytables seen
-has Bool %.aoh-seen;
+has Bool:D %!aoh-seen{Array:D};
 
 # TOML table tracker, records tables seen
-has Bool %.hoh-seen;
+has Bool:D %!hoh-seen{Array:D};
 
 # TOML key tracker, records keypair keys seen
-has Bool %.keys-seen;
+has Bool:D %!keys-seen{Array:D};
 
 # DateTime offset for when the local offset is omitted in TOML dates,
 # see: https://github.com/toml-lang/toml#datetime
 # if not passed as a parameter during instantiation, use host machine's
 # local offset
-has Int $.date-local-offset = $*TZ;
+has Int:D $.date-local-offset = $*TZ;
 
 # string grammar-actions {{{
 
@@ -71,12 +72,12 @@ method escape:sym<backslash>($/)
 
 method escape:sym<u>($/)
 {
-    make chr :16(@<hex>.join);
+    make chr(:16(@<hex>.join));
 }
 
 method escape:sym<U>($/)
 {
-    make chr :16(@<hex>.join);
+    make chr(:16(@<hex>.join));
 }
 
 method string-basic-char:escape-sequence ($/)
@@ -213,6 +214,16 @@ method float($/)
     make +$/;
 }
 
+method plus-or-minus:sym<+>($/)
+{
+    make ~$/;
+}
+
+method plus-or-minus:sym<->($/)
+{
+    make ~$/;
+}
+
 method number($/)
 {
     if $<integer>
@@ -278,7 +289,7 @@ method time-secfrac($/)
 
 method time-numoffset($/)
 {
-    my Int $multiplier = $<plus-or-minus> ~~ '+' ?? 1 !! -1;
+    my Int:D $multiplier = $<plus-or-minus>.made eq '+' ?? 1 !! -1;
     make Int(
         (
             ($multiplier * $<time-hour>.made * 60) + $<time-minute>.made
@@ -294,20 +305,12 @@ method time-offset($/)
 
 method partial-time($/)
 {
-    my Rat $second = Rat($<time-second>.made);
-    my Bool $subseconds = False;
-
-    if $<time-secfrac>
-    {
-        $second += Rat($<time-secfrac>.made);
-        $subseconds = True;
-    }
-
+    my Rat:D $second = Rat($<time-second>.made);
+    $second += Rat($<time-secfrac>.made) if $<time-secfrac>;
     make %(
         :hour(Int($<time-hour>.made)),
         :minute(Int($<time-minute>.made)),
-        :$second,
-        :$subseconds
+        :$second
     );
 }
 
@@ -326,14 +329,13 @@ method full-time($/)
         :hour(Int($<partial-time>.made<hour>)),
         :minute(Int($<partial-time>.made<minute>)),
         :second(Rat($<partial-time>.made<second>)),
-        :subseconds(Bool($<partial-time>.made<subseconds>)),
         :timezone(Int($<time-offset>.made))
     );
 }
 
 method date-time-omit-local-offset($/)
 {
-    make DateTime.new(
+    make %(
         :year(Int($<full-date>.made<year>)),
         :month(Int($<full-date>.made<month>)),
         :day(Int($<full-date>.made<day>)),
@@ -346,7 +348,7 @@ method date-time-omit-local-offset($/)
 
 method date-time($/)
 {
-    make DateTime.new(
+    make %(
         :year(Int($<full-date>.made<year>)),
         :month(Int($<full-date>.made<month>)),
         :day(Int($<full-date>.made<day>)),
@@ -359,17 +361,17 @@ method date-time($/)
 
 method date:full-date ($/)
 {
-    make DateTime.new(|$<full-date>.made, :timezone($.date-local-offset));
+    make Date.new(|$<full-date>.made);
 }
 
 method date:date-time-omit-local-offset ($/)
 {
-    make $<date-time-omit-local-offset>.made;
+    make DateTime.new(|$<date-time-omit-local-offset>.made);
 }
 
 method date:date-time ($/)
 {
-    make $<date-time>.made;
+    make DateTime.new(|$<date-time>.made);
 }
 
 # end datetime grammar-actions }}}
@@ -463,22 +465,14 @@ method keypair-value:array ($/)
     make $<array>.made;
 }
 
+method keypair-value:table-inline ($/)
+{
+    make $<table-inline>.made;
+}
+
 method keypair($/)
 {
-    # if keypair value is inline table, map inline table keypairs to
-    # keypair key as hash
-    if $<table-inline>
-    {
-        my %h;
-        $<table-inline>.made.map({
-            %h{Str($<keypair-key>.made)}{.keys[0]} = .values[0]
-        });
-        make %h;
-    }
-    else
-    {
-        make Str($<keypair-key>.made) => $<keypair-value>.made;
-    }
+    make Str($<keypair-key>.made) => $<keypair-value>.made;
 }
 
 method table-inline-keypairs($/)
@@ -489,12 +483,13 @@ method table-inline-keypairs($/)
     # method being assigned in a hash and are at risk of being overwritten
     # by duplicate keys
     {
-        my Str @keys-seen = |@<keypair>».made».keys.flat;
+        my Str:D @keys-seen = |@<keypair>».made».keys.flat;
         unless @keys-seen.elems == @keys-seen.unique.elems
         {
             die X::Config::TOML::InlineTable::DuplicateKeys.new(
-                :table-inline-text($/.Str),
-                :@keys-seen
+                :@keys-seen,
+                :subject('inline table'),
+                :text(~$/)
             );
         }
     }
@@ -529,19 +524,26 @@ method keypair-line($/)
 # this segment represents keypairs not belonging to any table
 method segment:keypair-line ($/)
 {
-    # update keypath
-    my Str @keypath = $<keypair-line>.made.keys[0];
+    my @path = $<keypair-line>.made.keys[0];
+    my $value = $<keypair-line>.made.values[0];
 
-    # mark key as seen, verify key is not being redeclared
-    if %!keys-seen{$(self!pwd(%.toml, @keypath))}++
+    if seen(%!keys-seen, :@path)
     {
         die X::Config::TOML::KeypairLine::DuplicateKeys.new(
-            :keypair-line-text($/.Str),
-            :@keypath
+            :keypair-line-text(~$/),
+            :@path
         );
     }
 
-    self!at-keypath(%!toml, @keypath) = $<keypair-line>.made.values[0];
+    Crane.exists(%!toml, :@path)
+        ?? die
+            X::Config::TOML::KeypairLine::DuplicateKeys.new(
+                :keypair-line-text(~$/),
+                :@path
+            )
+        !! Crane.set(%!toml, :@path, :$value);
+
+    %!keys-seen{$@path}++;
 }
 
 method table-header-text($/)
@@ -556,108 +558,94 @@ method hoh-header($/)
 
 method table:hoh ($/)
 {
-    my Str @base-keypath = $<hoh-header>.made;
-    my Str $hoh-text = $/.Str;
+    my @base-path = pwd(%!toml, :steps($<hoh-header>.made));
+    my Str:D $hoh-text = ~$/;
 
-    # verify table does not overwrite existing key
-    if %.keys-seen{$(self!pwd(%.toml, @base-keypath))}
+    if seen(%!keys-seen, :path(@base-path))
     {
         die X::Config::TOML::HOH::Seen::Key.new(
             :$hoh-text,
-            :keypath(@base-keypath)
+            :path(@base-path)
         );
     }
-
-    # verify table does not overwrite existing arraytable
-    if %.aoh-seen{$@base-keypath}
+    if %!aoh-seen.grep({.keys[0] eqv $@base-path}).elems > 0
     {
         die X::Config::TOML::HOH::Seen::AOH.new(
-            :hoh-header-text($<hoh-header>.Str),
+            :hoh-header-text(~$<hoh-header>),
             :$hoh-text
         );
     }
-
-    # mark table as seen, verify table is not being redeclared
-    if %!hoh-seen{$(self!pwd(%.toml, @base-keypath))}++
+    if %!hoh-seen.grep({.keys[0] eqv $@base-path}).elems > 0
     {
         die X::Config::TOML::HOH::Seen.new(
-            :hoh-header-text($<hoh-header>.Str),
+            :hoh-header-text(~$<hoh-header>),
             :$hoh-text
         );
     }
 
-    # verify base keypath is clear
-    try
+    my @keypairs = @<keypair-line>».made.flat;
     {
-        self.is-keypath-clear(@base-keypath);
-
         CATCH
         {
-            default
+            when X::AdHoc
             {
-                die X::Config::TOML::Keypath::HOH.new(
-                    :$hoh-text,
-                    :keypath(@base-keypath)
-                );
-            }
-        }
-    }
-
-    # does table contain keypairs?
-    if @<keypair-line>
-    {
-        # verify keypair lines do not contain duplicate keys
-        {
-            my Str @keys-seen = |@<keypair-line>».made».keys.flat;
-            unless @keys-seen.elems == @keys-seen.unique.elems
-            {
-                die X::Config::TOML::HOH::DuplicateKeys.new(
-                    :$hoh-text,
-                    :@keys-seen
-                );
-            }
-        }
-
-        for @<keypair-line>».made -> $keypair
-        {
-            my Str @keypath = @base-keypath;
-            push @keypath, $keypair.keys[0];
-
-            # verify keypair key does not conflict with existing key
-            try
-            {
-                self.is-keypath-clear(@keypath);
-
-                CATCH
+                my rule exception-associative-indexing
                 {
-                    when X::Config::TOML::BadKeypath::ArrayNotAOH
-                    {
-                        die X::Config::TOML::HOH::Seen::Key.new(
-                            :$hoh-text,
-                            :keypath(@base-keypath)
-                        );
-                    }
-                    default
-                    {
-                        die X::Config::TOML::Keypath::HOH.new(
-                            :$hoh-text,
-                            :@keypath
-                        );
-                    }
+                    Type (\w+) does not support associative indexing
+                }
+                if .payload ~~ &exception-associative-indexing
+                {
+                    die X::Config::TOML::HOH::Seen::Key.new(
+                        :$hoh-text,
+                        :path(@base-path)
+                    );
+                }
+            }
+        }
+
+        # does table contain keypairs?
+        if @keypairs
+        {
+            # verify keypairs do not contain duplicate keys
+            {
+                my Str:D @keys-seen = |@keypairs».keys.flat;
+                unless @keys-seen.elems == @keys-seen.unique.elems
+                {
+                    die X::Config::TOML::HOH::DuplicateKeys.new(
+                        :@keys-seen,
+                        :subject('table'),
+                        :text($hoh-text)
+                    );
                 }
             }
 
-            # assign value to keypath
-            self!at-keypath(%!toml, @keypath) = $keypair.values[0];
-
-            # mark key as seen
-            %!keys-seen{$(self!pwd(%.toml, @keypath))}++;
+            for @keypairs -> %keypair
+            {
+                my @path = |@base-path, %keypair.keys[0];
+                my $value = %keypair.values[0];
+                Crane.exists(%!toml, :@path)
+                    ?? die
+                        X::Config::TOML::HOH::Seen::Key.new(
+                            :$hoh-text,
+                            :@path
+                        )
+                    !! Crane.set(%!toml, :@path, :$value);
+                %!keys-seen{$@path}++;
+            }
+        }
+        else
+        {
+            Crane.exists(%!toml, :path(@base-path))
+                ?? die
+                    X::Config::TOML::HOH::Seen::Key.new(
+                        :$hoh-text,
+                        :path(@base-path)
+                    )
+                !! Crane.set(%!toml, :path(@base-path), :value({}));
         }
     }
-    else
-    {
-        self!at-keypath(%!toml, @base-keypath) = {};
-    }
+
+    %!hoh-seen{$@base-path}++;
 }
 
 method aoh-header($/)
@@ -667,287 +655,127 @@ method aoh-header($/)
 
 method table:aoh ($/)
 {
-    my Str @base-keypath = $<aoh-header>.made;
-    my Str $aoh-header-text = $<aoh-header>.Str;
-    my Str $aoh-text = $/.Str;
+    my @path = pwd(%!toml, :steps($<aoh-header>.made));
+    my Str:D $aoh-header-text = ~$<aoh-header>;
+    my Str:D $aoh-text = ~$/;
 
-    # verify arraytable does not overwrite existing key
-    if %.keys-seen{$(self!pwd(%.toml, @base-keypath))}
+    if seen(%!keys-seen, :@path)
     {
         die X::Config::TOML::AOH::OverwritesKey.new(
             :$aoh-header-text,
             :$aoh-text,
-            :keypath(@base-keypath)
+            :@path
         );
     }
-
-    # verify arraytable does not overwrite existing table
-    if %.hoh-seen{$(self!pwd(%.toml, @base-keypath))}
+    if %!hoh-seen.grep({.keys[0] eqv $@path}).elems > 0
     {
         die X::Config::TOML::AOH::OverwritesHOH.new(
             :$aoh-header-text,
             :$aoh-text,
-            :keypath(@base-keypath)
+            :@path
         );
     }
 
+    unless %!aoh-seen.grep({.keys[0] eqv $@path}).elems > 0
+    {
+        Crane.exists(%!toml, :@path)
+            ?? die X::Config::TOML::Keypath::AOH.new(:$aoh-text, :@path)
+            !! Crane.set(%!toml, :@path, :value([]));
+        %!aoh-seen{$@path}++;
+    }
+
     my %h;
-    if @<keypair-line>
+    my @keypairs = @<keypair-line>».made.flat;
+    if @keypairs
     {
         # verify keypair lines do not contain duplicate keys
         {
-            my Str @keys-seen = |@<keypair-line>».made».keys.flat;
+            my Str:D @keys-seen = |@keypairs».keys.flat;
             unless @keys-seen.elems == @keys-seen.unique.elems
             {
                 die X::Config::TOML::AOH::DuplicateKeys.new(
-                    :$aoh-text,
-                    :@keys-seen
+                    :@keys-seen,
+                    :subject('array table'),
+                    :text($aoh-text)
                 );
             }
         }
 
-        @<keypair-line>».made.map({ %h{.keys[0]} = .values[0]; });
-    }
-
-    sub append-to-aoh(@keypath, %h)
-    {
-        push self!at-keypath(
-            %!toml,
-            @keypath.end > 0 ?? @keypath[0..^@keypath.end] !! []
-        ){@keypath[@keypath.end]}, %h;
-    }
-
-    # is base keypath an existing array of hashes?
-    if %.aoh-seen{$@base-keypath}
-    {
-        # push values to existing array of hashes
-        append-to-aoh(@base-keypath, %h);
-    }
-    # new array of hashes
-    else
-    {
-        # make sure we're not trodding over scalars or tables
-        try
-        {
-            self.is-keypath-clear(@base-keypath, :aoh);
-
-            CATCH
-            {
-                when X::Config::TOML::BadKeypath::ArrayNotAOH
-                {
-                    die X::Config::TOML::AOH::OverwritesKey.new(
-                        :$aoh-header-text,
-                        :$aoh-text,
-                        :keypath(@base-keypath)
-                    );
-                }
-                default
-                {
-                    die X::Config::TOML::Keypath::AOH.new(
-                        :$aoh-text,
-                        :keypath(@base-keypath)
-                    );
-                }
-            }
-        }
-
-        # push values to new array of hashes
-        append-to-aoh(@base-keypath, %h);
-
-        # mark arraytable as seen
-        %!aoh-seen{$@base-keypath}++;
+        @keypairs.map({ %h{.keys[0]} = .values[0] });
     }
+    Crane.set(%!toml, :path(|@path, *-0), :value(%h));
 }
 
 method TOP($/)
 {
-    make %.toml;
+    make %!toml;
 }
 
 # end document grammar-actions }}}
 
 # helper functions {{{
 
-# given TOML hash and keypath, return scalar container of deepest path,
-# with special treatment of array of hashes
-method !at-keypath(%h, *@k) is rw
-{
-    my $h := %h;
-
-    my @path;
-
-    for @k -> $k
-    {
-        push @path, $k;
-
-        # if keypath step is array of hashes, always traverse array
-        # element of highest index
-        if $h{$k} ~~ List
-        {
-            # verify this is aoh before traversing
-            unless %.aoh-seen{$@path}
-            {
-                die X::Config::TOML::BadKeypath::ArrayNotAOH.new;
-            }
-
-            my Int $l = $h{$k}.end;
-            $h := $h{$k}[$l];
-        }
-        else
-        {
-            $h := $h{$k};
-        }
-    }
-    $h;
-}
-
-# verify keypath does not conflict with existing key
-multi method is-keypath-clear(Str:D @full-keypath) returns Bool:D
-{
-    my Bool:D $clear = False;
-
-    # does full keypath exist?
-    if self!at-keypath(%.toml, @full-keypath).defined
-    {
-        # is it a scalar?
-        unless self!at-keypath(%.toml, @full-keypath).WHAT ~~ Hash
-        {
-            $clear = False;
-            die X::Config::TOML::Keypath.new(:keypath(@full-keypath));
-        }
-
-        $clear = True;
-    }
-    else
-    {
-        # full keypath does not exist, and full keypath has depth of 1?
-        if @full-keypath.end == 0
-        {
-            $clear = True;
-        }
-        else
-        {
-            # for extended keypaths, make sure we're not trodding
-            # over scalars
-            $clear = self._is-keypath-clear(
-                @full-keypath[0..^@full-keypath.end].Array
-            );
-        }
-    }
-
-    $clear;
-}
-
-# special keypath check for arraytables
-multi method is-keypath-clear(
-    Str:D @full-keypath,
-    Bool:D :$aoh! where *.so
-) returns Bool:D
-{
-    my Bool:D $clear = False;
-
-    # does full keypath exist?
-    if self!at-keypath(%.toml, @full-keypath).defined
-    {
-        # we're tracking arraytables so if we got here, it's because
-        # we're tasked with making a new arraytable
-        # this new arraytable cannot overwrite any existing value
-        $clear = False;
-        die X::Config::TOML::Keypath.new(:keypath(@full-keypath));
-    }
-    else
-    {
-        # full keypath does not exist, and full keypath has depth of 1?
-        if @full-keypath.end == 0
-        {
-            $clear = True;
-        }
-        else
-        {
-            # for extended keypaths, make sure we're not trodding
-            # over scalars
-            $clear = self._is-keypath-clear(
-                @full-keypath[0..^@full-keypath.end].Array
-            );
-        }
-    }
-
-    $clear;
-}
-
-multi method _is-keypath-clear(@keypath where *.end > 0) returns Bool:D
-{
-    self!is-trodden(@keypath)
-        ?? die X::Config::TOML::Keypath.new(:@keypath)
-        !! self._is-keypath-clear(@keypath[0..^@keypath.end]);
-}
-
-multi method _is-keypath-clear(@keypath where *.end == 0) returns Bool:D
-{
-    self!is-trodden(@keypath)
-        ?? die X::Config::TOML::Keypath.new(:@keypath)
-        !! True;
-}
-
-method !is-trodden(@keypath) returns Bool:D
-{
-    if self!at-keypath(%.toml, @keypath).defined
-    {
-        unless self!at-keypath(%.toml, @keypath).WHAT ~~ Hash
-        {
-            die X::Config::TOML::Keypath.new(:@keypath);
-            True;
-        }
-    }
-    False;
-}
-
 # given TOML hash and keypath, print working directory including
 # arraytable indices
-# returns either a keyname or array index at each step of the path
-class Step { has $.key; has Int $.index; }
-method !pwd(%h, *@k) returns Array
+multi sub pwd(Associative:D $container, :@steps where *.elems > 0) returns Array:D
 {
-    my Step @steps;
-
-    my $h := %h;
-    for @k -> $k
-    {
-        # if keypath step is array of hashes, always traverse array
-        # element of highest index
-        if $h{$k} ~~ List
-        {
-            my Int $l = $h{$k}.end;
-            $h := $h{$k}[$l];
-            push @steps, Step.new(:key($k), :index($l));
-        }
-        else
-        {
-            $h := $h{$k};
-            push @steps, Step.new(:key($k));
-        }
-    }
-
-    unfold(@steps);
+    my @steps-taken;
+    my $root := $container;
+    $root := $root{@steps[0]};
+    push @steps-taken, @steps[0], |pwd($root, :steps(@steps[1..*]));
+    @steps-taken;
 }
 
-# convert list of Steps to strings
-sub unfold(Step @steps) returns Array
+multi sub pwd(Associative:D $container, :@steps where *.elems == 0) returns Array:D
 {
-    my Str @unfold;
+    my @steps-taken;
+}
 
-    for @steps -> $step
-    {
-        # surround keypair keys with angle brackets
-        push @unfold, '<' ~ $step.key ~ '>';
+multi sub pwd(Positional:D $container, :@steps where *.elems > 0) returns Array:D
+{
+    my @steps-taken;
+    my $root := $container;
+    my Int:D $index = $container.end;
+    $root := $root[$index];
+    push @steps-taken, $index, |pwd($root, :@steps);
+    @steps-taken;
+}
 
-        # surround array indices with square brackets
-        push @unfold, '[' ~ $step.index ~ ']' if defined $step.index;
-    }
+multi sub pwd(Positional:D $container, :@steps where *.elems == 0) returns Array:D
+{
+    my @steps-taken;
+}
 
-    @unfold;
+multi sub pwd($container, :@steps where *.elems > 0) returns Array:D
+{
+    my @steps-taken;
+    my $root := $container;
+    $root := $root{@steps[0]};
+    push @steps-taken, @steps[0], |pwd($root, :steps(@steps[1..*]));
+    @steps-taken;
+}
+
+multi sub pwd($container, :@steps where *.elems == 0) returns Array:D
+{
+    my @steps-taken;
+}
+
+multi sub seen(Bool:D %h, :@path! where *.elems > 1) returns Bool:D
+{
+    %h.grep({.keys[0] eqv $@path}).elems > 0
+        || seen(%h, :path(@path[0..^*-1].Array));
+}
+
+multi sub seen(Bool:D %h, :@path! where *.elems > 0) returns Bool:D
+{
+    %h.grep({.keys[0] eqv $@path}).elems > 0;
+}
+
+multi sub seen(Bool:D %h, :@path! where *.elems == 0) returns Bool:D
+{
+    False;
 }
 
 # end helper functions }}}
 
-# vim: ft=perl6 fdm=marker fdl=0
+# vim: set filetype=perl6 foldmethod=marker foldlevel=0:
