@@ -1,11 +1,12 @@
 use v6;
+use X::Config::TOML;
 unit grammar Config::TOML::Parser::Grammar;
 
 # disposable grammar {{{
 
 proto token gap {*}
 token gap:spacer { \s }
-token gap:comment { <.comment> \n }
+token gap:comment { <.comment> $$ }
 
 # end disposable grammar }}}
 # comment grammar {{{
@@ -66,8 +67,7 @@ token string-basic-char:escape-sequence
 
         .
         {
-            say "Sorry, found bad TOML escape sequence 「$/」";
-            exit;
+            die X::Config::TOML::String::EscapeSequence.new(:esc(~$/));
         }
     ]
 }
@@ -157,8 +157,7 @@ token string-basic-multiline-char:escape-sequence
 
         .
         {
-            say "Sorry, found bad TOML escape sequence 「$/」";
-            exit;
+            die X::Config::TOML::String::EscapeSequence.new(:esc(~$/));
         }
     ]
 }
@@ -516,7 +515,7 @@ token table-inline-keypairs
 
 token keypair
 {
-    <keypair-key> \h* '=' \h* [ <keypair-value> | <table-inline> ]
+    <keypair-key> \h* '=' \h* <keypair-value>
 }
 
 proto token keypair-key {*}
@@ -535,6 +534,7 @@ token keypair-value:number { <number> }
 token keypair-value:boolean { <boolean> }
 token keypair-value:date { <date> }
 token keypair-value:array { <array> }
+token keypair-value:table-inline { <table-inline> }
 
 # end table grammar }}}
 # document grammar {{{
@@ -542,19 +542,19 @@ token keypair-value:array { <array> }
 # blank line
 token blank-line
 {
-    ^^ \h* $$ \n
+    ^^ \h* $$
 }
 
 # comment appearing on its own line
 token comment-line
 {
-    ^^ \h* <.comment> $$ \n
+    ^^ \h* <.comment> $$
 }
 
 # keypair appearing on its own line
 token keypair-line
 {
-    ^^ \h* <keypair> \h* <.comment>? $$ \n
+    ^^ \h* <keypair> \h* <.comment>? $$
 }
 
 proto token table {*}
@@ -562,15 +562,19 @@ proto token table {*}
 # standard TOML table (hash of hashes)
 token table:hoh
 {
-    ^^ \h* <hoh-header> \h* <.comment>? $$ \n
-    [ <keypair-line> | <.comment-line> | <.blank-line> ]*
+    ^^ \h* <hoh-header> \h* <.comment>? $$
+    [
+        \n [ <keypair-line> | <.comment-line> | <.blank-line> ]
+    ]*
 }
 
 # TOML array of tables (array of hashes)
 token table:aoh
 {
-    ^^ \h* <aoh-header> \h* <.comment>? $$ \n
-    [ <keypair-line> | <.comment-line> | <.blank-line> ]*
+    ^^ \h* <aoh-header> \h* <.comment>? $$
+    [
+        \n [ <keypair-line> | <.comment-line> | <.blank-line> ]
+    ]*
 }
 
 # hash of hashes header
@@ -614,7 +618,13 @@ token segment:table
 
 token document
 {
-    <segment>*
+    [
+        <segment>
+        [
+            \n <segment>
+        ]*
+    ]?
+    \n?
 }
 
 token TOP
@@ -624,4 +634,4 @@ token TOP
 
 # end document grammar }}}
 
-# vim: ft=perl6 fdm=marker fdl=0
+# vim: set filetype=perl6 foldmethod=marker foldlevel=0:
